local em = GetEventManager()
local _
local db, tlw
local dx = 1/GetSetting(SETTING_TYPE_UI, UI_SETTING_CUSTOM_SCALE) --Get UI Scale to draw thin lines correctly
PENTEST_UI_SCALE = dx
local BOLD_FONT = "EsoUI/Common/Fonts/Univers67.otf"

-- Addon Namespace
PenTest = PenTest or {}
local PenTest = PenTest
PenTest.name 		= "PenTest"
PenTest.version 	= "26"
PenTest.debug = false

-- Logger

local mainlogger
local LOG_LEVEL_VERBOSE = "V"
local LOG_LEVEL_DEBUG = "D"
local LOG_LEVEL_INFO = "I"
local LOG_LEVEL_WARNING ="W"
local LOG_LEVEL_ERROR = "E"

if LibDebugLogger then

	mainlogger = LibDebugLogger.Create(PenTest.name)

	LOG_LEVEL_VERBOSE = LibDebugLogger.LOG_LEVEL_VERBOSE
	LOG_LEVEL_DEBUG = LibDebugLogger.LOG_LEVEL_DEBUG
	LOG_LEVEL_INFO = LibDebugLogger.LOG_LEVEL_INFO
	LOG_LEVEL_WARNING = LibDebugLogger.LOG_LEVEL_WARNING
	LOG_LEVEL_ERROR = LibDebugLogger.LOG_LEVEL_ERROR

end

local function Print(level, ...)

	if mainlogger == nil then return end

	local logger = mainlogger

	if type(logger.Log)=="function" then logger:Log(level, ...) end

end

-- Data for tracked abilities.
local PENTEST_TYPE_SPELLPEN = 1
local PENTEST_TYPE_WEAPONPEN = 2
local PENTEST_TYPE_ALLPEN = 3
local PENTEST_TYPE_VULNERABILITY = 4
local PENTEST_TYPE_CRITBONUS = 5
local PENTEST_TYPE_OFFBALANCE = 6

local PENTEST_OFFBALANCE = 1
local PENTEST_OFFBALANCE_IMMUNE = 2

local AbilityData
local backstabberBaseValue

local function GetAbilityData()

	AbilityData = {

		[GetAbilityName(62787)] = { 			-- Major Breach
			value = 5948,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[GetAbilityName(68588)] = { 			-- Minor Breach
			value = 2974,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[17906] = {								-- Crusher, can get changed by settings !
			value = db.crusherValue or 2108,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[120007] = {							-- Crusher, Trial Dummy
			value = 2108,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[75753] = {								-- Alkosh (can be lower)
			value = 6000,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[120018] = {							-- Alkosh, Trial Dummy
			value = 6000,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[GetAbilityName(143808)] = {			-- Crystal Weapon (Sorc)
			value = 1000,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[GetAbilityName(159288)] = {			-- Crimson Oath
			value = 3541,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[GetAbilityName(79087)] = {				-- Spell Resistance Reduction by Poison
			value = 1320,
			datatype = PENTEST_TYPE_SPELLPEN,
		},

		[GetAbilityName(79090)] = {				-- Physical Resistance Reduction by Poison
			value = 1320,
			datatype = PENTEST_TYPE_WEAPONPEN,
		},

		[GetAbilityName(80866)] = {				-- Tremorscale
			value = 2640,
			datatype = PENTEST_TYPE_ALLPEN,
		},

		[GetAbilityName(39077)] = {				-- Off Balance
			value = PENTEST_OFFBALANCE,
			datatype = PENTEST_TYPE_OFFBALANCE,
		},

		[GetAbilityName(134599)] = {			-- Off Balance Immunity
			value = PENTEST_OFFBALANCE_IMMUNE,
			datatype = PENTEST_TYPE_OFFBALANCE,
		},

		[GetAbilityName(68359)] = {				-- Minor Vulnerabiltiy (Concussion)
			value = 5,
			datatype = PENTEST_TYPE_VULNERABILITY,
		},

		[GetAbilityName(81519)] = {				-- Minor Vulnerabiltiy (Infallible Aether)
			value = 5,
			datatype = PENTEST_TYPE_VULNERABILITY,
		},

		[GetAbilityName(142610)] = {			-- Flame Weakness
			value = 5,
			datatype = PENTEST_TYPE_CRITBONUS,
		},

		[GetAbilityName(142653)] = {			-- Shock Weakness
			value = 5,
			datatype = PENTEST_TYPE_CRITBONUS,
		},

		[GetAbilityName(142652)] = {			-- Frost Weakness
			value = 5,
			datatype = PENTEST_TYPE_CRITBONUS,
		},

		[GetAbilityName(145975)] = {			-- Minor Brittle
			value = 10,
			datatype = PENTEST_TYPE_CRITBONUS,
		},

		[GetAbilityName(61723)] = {				-- Minor Maim (Chilled) in combination with the Warden passive "Glacial Presence"
			value = 0,								-- Needs to be zero since it doesn't add crit damage on its own!
			datatype = PENTEST_TYPE_CRITBONUS,
		},

	}	--Corrosive Armor ignores all resistance

	backstabberBaseValue = 2
end

local function GetCleanDimensions(x, y)

	if type(x) == "number" then x = math.max(zo_round(x)*dx - 0.02, 0) end
	if type(y) == "number" then y = math.max(zo_round(y)*dx - 0.02, 0) end

	return x, y

end

local function UpdateLayout()

	local dbtlw = db.tlw

	local BG = tlw:GetNamedChild("Bg")
	BG:SetAlpha(db.opacity/100)

	local penControl = tlw:GetNamedChild("PenetrationControl")

	local width = dbtlw.width
	local height = dbtlw.height

	local spellPenControl = penControl:GetNamedChild("Spell")

	if db.spellpen then

		spellPenControl:SetHidden(false)
		spellPenControl:SetDimensions(GetCleanDimensions(width + 4 , height + 8))

		local penBarS = spellPenControl:GetNamedChild("Bar")
		penBarS:SetDimensions(GetCleanDimensions(width, height))
		penBarS:SetCenterColor(0.27, 0.27, 0.6, 1)

		local penBgS = spellPenControl:GetNamedChild("Bg")
		penBgS:SetDimensions(GetCleanDimensions(width, height))

		local penLabelS = spellPenControl:GetNamedChild("Label")
		penLabelS:SetFont(string.format("%s|%d|soft-shadow-thin", BOLD_FONT, height*.75 ))

	else

		spellPenControl:SetHidden(true)

	end

	local weaponPenControl = penControl:GetNamedChild("Weapon")

	if db.weaponpen then

		weaponPenControl:SetHidden(false)
		weaponPenControl:SetDimensions(GetCleanDimensions(width + 4 , height + 8))

		local penBarW = weaponPenControl:GetNamedChild("Bar")
		penBarW:SetDimensions(GetCleanDimensions(width, height))
		penBarW:SetCenterColor(0.33, 0.6, 0.27, 1)

		local penBgW = weaponPenControl:GetNamedChild("Bg")
		penBgW:SetDimensions(GetCleanDimensions(width, height))

		local penLabelW = weaponPenControl:GetNamedChild("Label")
		penLabelW:SetFont(string.format("%s|%d|soft-shadow-thin", BOLD_FONT, height *.75 ))

	else

		weaponPenControl:SetHidden(true)

	end

	local size = db.iconsize
	local iconControl = tlw:GetNamedChild("IconsControl")
	-- iconControl:SetDimensions(GetCleanDimensions((size*3)+8, size + 8))

	local icon1 = iconControl:GetNamedChild("CritDamageIcon")
	local icon2 = iconControl:GetNamedChild("OffBalanceIcon")
	local icon3 = iconControl:GetNamedChild("VulnerabilityIcon")

	local anchor = {LEFT, iconControl, LEFT, 0, 0}

	if db.CritDamageIcon then

		icon1:SetDimensions(GetCleanDimensions(size, size))
		icon1:SetHidden(false)

		local label = icon1:GetNamedChild("Label")
		label:SetFont(string.format("%s|%d|soft-shadow-thin", BOLD_FONT, size *.50 ))
		anchor = {LEFT, icon1, RIGHT, 4, 0}

	else

		icon1:SetHidden(true)

	end

	if db.OffBalanceIcon then

		icon2:SetDimensions(GetCleanDimensions(size, size))
		icon2:ClearAnchors()
		icon2:SetAnchor(unpack(anchor))		-- Update Anchors
		icon2:SetHidden(false)

		local label2 = icon2:GetNamedChild("Label")
		label2:SetFont(string.format("%s|%d|soft-shadow-thin", BOLD_FONT, size *.7 ))

		anchor = {LEFT, icon2, RIGHT, 4, 0}

	else

		icon2:SetHidden(true)

	end

	if db.MinorVulnerabilityIcon then

		icon3:SetDimensions(GetCleanDimensions(size, size))
		icon3:ClearAnchors()
		icon3:SetAnchor(unpack(anchor))		-- Update Anchors
		icon3:SetHidden(false)

		local label3 = icon3:GetNamedChild("Label")
		label3:SetFont(string.format("%s|%d|soft-shadow-thin", BOLD_FONT, size *.7 ))

	else

		icon3:SetHidden(true)

	end

	tlw:SetResizeToFitDescendents(false)
	penControl:SetResizeToFitDescendents(false)
	iconControl:SetResizeToFitDescendents(false)

	tlw:SetResizeToFitDescendents(true)
	penControl:SetResizeToFitDescendents(true)
	iconControl:SetResizeToFitDescendents(true)

end

local red = ZO_ColorDef:New(1, 0, 0, 1)
local yellow = ZO_ColorDef:New(1, 1, 0, 1)
local green = ZO_ColorDef:New(0, 0.9, 0, 1)
local white = ZO_ColorDef:New(1, 1, 1, 1)

local function UpdateControl(spellpen, weaponpen, vulnerability, offBalanceStatus, offBalanceTime, vulnTime, critbonus)

	local maxwidth = db.tlw.width
	local maxpen = db.maxpen

	local penControl = tlw:GetNamedChild("PenetrationControl")

	local spellPenControl = penControl:GetNamedChild("Spell")

	local penBar = spellPenControl:GetNamedChild("Bar")
	local penLabel = spellPenControl:GetNamedChild("Label")

	local sizeS = math.min(spellpen / maxpen, 1)
	local sizeW = math.min(weaponpen / maxpen, 1)

	local widthS, widthW = GetCleanDimensions(sizeS * maxwidth, sizeW * maxwidth)

	local textcolorS = spellpen >= maxpen + 600 and red or spellpen >= maxpen - 600 and green or white
	local textcolorW = weaponpen >= maxpen + 600 and red or weaponpen >= maxpen - 600 and green or white

	penBar:SetWidth(widthS)
	penLabel:SetText(spellpen)
	penLabel:SetColor(textcolorS:UnpackRGBA())

	local weaponPenControl = penControl:GetNamedChild("Weapon")

	local penBar2 = weaponPenControl:GetNamedChild("Bar")
	local penLabel2 = weaponPenControl:GetNamedChild("Label")

	penBar2:SetWidth(widthW)
	penLabel2:SetText(weaponpen)
	penLabel2:SetColor(textcolorW:UnpackRGBA())

	local iconControl = tlw:GetNamedChild("IconsControl")

	local critcolor = critbonus > 130 and red or critbonus > 125 and yellow or critbonus >= 120 and green or white

	local icon = iconControl:GetNamedChild("CritDamageIcon")
	icon:SetHidden((not db.CritDamageIcon))
	icon:GetNamedChild("Bg"):SetEdgeColor(critcolor:UnpackRGBA())

	local label = icon:GetNamedChild("Label")

	label:SetColor(critcolor:UnpackRGBA())
	label:SetText(string.format("%d%%", critbonus))

	local offBalanceIconId = offBalanceStatus == 2 and 134599 or 39077
	local offBalanceColor = offBalanceStatus == 1 and red or white

	local icon2 = iconControl:GetNamedChild("OffBalanceIcon")
	icon2:SetHidden((not db.OffBalanceIcon) or offBalanceStatus < 1)
	icon2:GetNamedChild("Texture"):SetTexture(GetAbilityIcon(offBalanceIconId))
	icon2:GetNamedChild("Bg"):SetEdgeColor(offBalanceColor:UnpackRGBA())

	local label2 = icon2:GetNamedChild("Label")

	label2:SetColor(offBalanceColor:UnpackRGBA())
	label2:SetText(string.format("%.1f", offBalanceTime))

	local icon3 = iconControl:GetNamedChild("VulnerabilityIcon")
	icon3:SetHidden((not db.MinorVulnerabilityIcon) or vulnerability == 0)

	local label3 = icon3:GetNamedChild("Label")

	label3:SetText(string.format("%.1f", vulnTime))

end

local staticValues = {

	backstabber = 0,
	critBonusMundus = 0,
	glacialPresence = 0,
	forceOfNature = 0

}

local DivineSlots = {EQUIP_SLOT_HEAD, EQUIP_SLOT_SHOULDERS, EQUIP_SLOT_CHEST, EQUIP_SLOT_HAND, EQUIP_SLOT_WAIST, EQUIP_SLOT_LEGS, EQUIP_SLOT_FEET}

local function GetStaticValues()

	local championBarData = CHAMPION_PERKS.championBar.slots
	local backstabberValue = CHAMPION_DATA_MANAGER.championSkillDataById[31]:GetNumSavedPoints()
	local forceOfNatureValue = CHAMPION_DATA_MANAGER.championSkillDataById[276]:GetNumSavedPoints()

	staticValues.backstabber = 0
	staticValues.forceOfNature = 0
	staticValues.critBonusMundus = 0
	staticValues.glacialPresence = 0

	-- Backstabber

	if backstabberValue > 0 or forceOfNatureValue > 0 then

		for i, slot in pairs(championBarData) do

			local slotData = slot:GetSavedChampionSkillData()

			if slotData and slotData:GetId() == 31 then staticValues.backstabber = backstabberBaseValue * math.floor(0.1 * backstabberValue) end
			if slotData and slotData:GetId() == 276 then staticValues.forceOfNature = db.forceOfNatureValue end

		end
	end

	-- Shadow Mundus (since ZOS does track it wrong)

	local effectSlot

	for i = 1, GetNumBuffs("player") do

		local _, _, _, buffSlot, _, _, _, _, _, _, abilityId, _, _ = GetUnitBuffInfo("player", i)

		if abilityId == 13984 then

			effectSlot = buffSlot
			break

		end
	end

	if effectSlot then

		local totalBonus = 0

		for _, key in pairs(DivineSlots) do

			local trait, desc = GetItemLinkTraitInfo(GetItemLink(BAG_WORN, key, LINK_STYLE_DEFAULT))

			if trait == ITEM_TRAIT_TYPE_ARMOR_DIVINES then

				local bonus = {desc:match("(%d+)%p?(%d*)[%%|]")}
				local bonusString = table.concat(bonus, ".")
				totalBonus = (tonumber(bonusString) or 0) + totalBonus

			end
		end

		local ZOSDesc = GetAbilityEffectDescription(effectSlot)
		local ZOSBonusString = ZOSDesc:match("cffffff(%d+)[%%|]")

		local calcBonus = math.floor(11 * (1 + totalBonus/100))
		local ZOSBonus = tonumber(ZOSBonusString) or 0 -- value attributed by ZOS

		staticValues.critBonusMundus = calcBonus - ZOSBonus -- mundus bonus difference

	end

	-- Glacial Presence Passive

	if GetUnitClassId("player") == 4 then

		local skillDataTable = SKILLS_DATA_MANAGER.abilityIdToProgressionDataMap
		local skillData = skillDataTable and skillDataTable[86192] and skillDataTable[86192].skillData

		local rank = skillData.currentRank	-- Glacial Presence
		local purchased = skillData.isPurchased	-- Glacial Presence

		local bonus = purchased and rank or 0

		staticValues.glacialPresence = bonus * 5

	end
end

local function GetCritbonus()

	local _, _, valueFromZos = GetAdvancedStatValue(ADVANCED_STAT_DISPLAY_TYPE_CRITICAL_DAMAGE)

	local total = 50 + valueFromZos + staticValues.backstabber

	return total

end

local function UpdateData()	-- acquire data on when reticle is over a target

	if not PenTest.inCombat then

		GetStaticValues()
		tlw:Toggle(not db.onlyCombat)

		if db.onlyCombat then return end

	else

		tlw:Toggle(true)

	end

	local baseSpellPen = GetPlayerStat(STAT_SPELL_PENETRATION, STAT_BONUS_OPTION_APPLY_BONUS)
	local baseWeaponPen = GetPlayerStat(STAT_PHYSICAL_PENETRATION, STAT_BONUS_OPTION_APPLY_BONUS)

	local sums = {0, 0, 0, 0, 0, 0}
	local offBalanceStatus = 0
	local offBalanceTime = 0
	local vulnTime = 0
	local correctionFactor = 0

	for i = 1, GetNumBuffs("player") do

		local buffName, timeStarted, timeEnding, buffSlot, stackCount, iconFilename, buffType, effectType, abilityType, statusEffectType, abilityId, canClickOff, castByPlayer = GetUnitBuffInfo("player", i)

		if abilityId == 51176 and stackCount > 1 then

			baseWeaponPen = baseWeaponPen + 544 * (stackCount - 1)
			break
			
		end
	end

	local glacialPresence = 0

	for buffIndex = 1,GetNumBuffs("reticleover") do

		local buffName, timeStarted, timeEnding, buffSlot, stackCount, iconFilename, buffType, effectType, abilityType, statusEffectType, abilityId, canClickOff, castByPlayer = GetUnitBuffInfo("reticleover", buffIndex)

		local data = AbilityData[GetAbilityName(abilityId)] or AbilityData[abilityId]

		if data then

			local datatype = data.datatype

			if datatype >= 1 and datatype <= 5 then

				glacialPresence = (abilityId == 145975 or abilityId == 61723) and staticValues.glacialPresence or 0	-- 61723 (Mainor Maim from Chilled) or 145975 (Minor Brittle)

				sums[datatype] = sums[datatype] + data.value

			elseif datatype == PENTEST_TYPE_OFFBALANCE then

				offBalanceStatus = data.value

				offBalanceTime = zo_roundToNearest(timeEnding - GetGameTimeMilliseconds()/1000, 0.1)

			end

			if datatype == PENTEST_TYPE_VULNERABILITY then vulnTime = zo_roundToNearest(timeEnding - GetGameTimeMilliseconds()/1000, 0.1) end
		end
	end

	local spellPen = sums[PENTEST_TYPE_SPELLPEN] + sums[PENTEST_TYPE_ALLPEN] + baseSpellPen + staticValues.forceOfNature
	local weaponPen = sums[PENTEST_TYPE_WEAPONPEN] + sums[PENTEST_TYPE_ALLPEN] + baseWeaponPen + staticValues.forceOfNature
	local vulnerability = sums[PENTEST_TYPE_VULNERABILITY]
	local critbonus = GetCritbonus() + sums[PENTEST_TYPE_CRITBONUS] + glacialPresence

	UpdateControl(spellPen, weaponPen, vulnerability, offBalanceStatus, offBalanceTime, vulnTime, critbonus)

end

local function onTargetChange()

	if DoesUnitExist("reticleover") == true and (GetUnitReaction("reticleover") == UNIT_REACTION_HOSTILE or GetUnitReaction("reticleover") == UNIT_REACTION_NEUTRAL) then

		EVENT_MANAGER:RegisterForUpdate("PenTest_Update", 100, UpdateData)
		UpdateData()

	else

		tlw:Toggle(false)
		EVENT_MANAGER:UnregisterForUpdate("PenTest_Update")

	end
end

local function onCombatState(event, inCombat)  -- called by Event

	PenTest.inCombat = inCombat

	if db.onlyCombat then tlw:Toggle(inCombat) end

	onTargetChange()
	GetAbilityData()
	GetStaticValues()

end

function PenTest.MakeMenu()
    -- load the settings->addons menu library
	local menu = LibAddonMenu2
	local def = PenTest.defaults

    -- the panel for the addons menu
	local panel = {
		type = "panel",
		name = "PenTest",
		displayName = "PenTest",
		author = "Solinur",
        version = PenTest.version or "",
		registerForRefresh = false,
	}

	local addonpanel = menu:RegisterAddonPanel("PenTest_Options", panel)

	local function ClearItems()

		if PenTest.inCombat then return end
		tlw:SetHidden(true)

	end

	local function ShowItems(currentpanel)

		if currentpanel ~= addonpanel then return end
		tlw:SetHidden(false)
		UpdateLayout()

	end

    --this adds entries in the addon menu
	local options = {
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_ACCOUNTWIDE),
			tooltip = GetString(SI_PENTEST_MENU_ACCOUNTWIDE_TOOLTIP),
			default = def.accountwide,
			getFunc = function() return PenTest_Save.Default[GetDisplayName()]['$AccountWide']["accountwide"] end,
			setFunc = function(value) PenTest_Save.Default[GetDisplayName()]['$AccountWide']["accountwide"] = value end,
			requiresReload = true,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_LOCK),
			tooltip = GetString(SI_PENTEST_MENU_LOCK_TOOLTIP),
			default = def.locked,
			getFunc = function() return db.locked end,
			setFunc = function(value)
						db.locked = value
						tlw:SetMovable(not value)
						tlw:SetMouseEnabled(not value)
					  end,
		},
		{
			type = "slider",
			name = GetString(SI_PENTEST_MENU_BG_OPACITY),
			tooltip = GetString(SI_PENTEST_MENU_BG_OPACITY_TOOLTIP),
			min = 0,
			max = 100,
			step = 5,
			default = def.opacity,
			getFunc = function() return db.opacity end,
			setFunc = function(value)
						db.opacity = value
						UpdateLayout()
					  end,
		},
		{
			type = "slider",
			name = GetString(SI_PENTEST_MENU_WINDOW_WIDTH),
			tooltip = GetString(SI_PENTEST_MENU_WINDOW_WIDTH_TOOLTIP),
			min = 100,
			max = 500,
			step = 10,
			default = def.tlw.width,
			getFunc = function() return zo_round(db.tlw.width) end,
			setFunc = function(value)
						db.tlw.width = zo_round(value/dx)*dx
						UpdateLayout()
					  end,
		},
		{
			type = "slider",
			name = GetString(SI_PENTEST_MENU_WINDOW_HEIGHT),
			tooltip = GetString(SI_PENTEST_MENU_WINDOW_HEIGHT_TOOLTIP),
			min = 15,
			max = 40,
			step = 1,
			default = def.tlw.height,
			getFunc = function() return zo_round(db.tlw.height) end,
			setFunc = function(value)
						db.tlw.height = zo_round(value/dx)*dx
						UpdateLayout()
					  end,
		},
		{
			type = "slider",
			name = GetString(SI_PENTEST_MENU_ICONSIZE),
			tooltip = GetString(SI_PENTEST_MENU_ICONSIZE_TOOLTIP),
			min = 15,
			max = 40,
			step = 1,
			default = def.iconsize,
			getFunc = function() return zo_round(db.iconsize) end,
			setFunc = function(value)
						db.iconsize = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_SHOW_SPELLPEN),
			tooltip = GetString(SI_PENTEST_MENU_SHOW_SPELLPEN_TOOLTIP),
			default = def.spellpen,
			getFunc = function() return db.spellpen end,
			setFunc = function(value)
						db.spellpen = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_SHOW_WEAPONPEN),
			tooltip = GetString(SI_PENTEST_MENU_SHOW_WEAPONPEN_TOOLTIP),
			default = def.weaponpen,
			getFunc = function() return db.weaponpen end,
			setFunc = function(value)
						db.weaponpen = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_SHOW_ICON):format(GetString(SI_PENTEST_CRIT_DAMAGE)),
			tooltip = GetString(SI_PENTEST_MENU_SHOW_ICON_TOOLTIP):format(GetString(SI_PENTEST_CRIT_DAMAGE)),
			default = def.CritDamageIcon,
			getFunc = function() return db.CritDamageIcon end,
			setFunc = function(value)
						db.CritDamageIcon = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_SHOW_ICON):format(GetString(SI_PENTEST_OFF_BALANCE)),
			tooltip = GetString(SI_PENTEST_MENU_SHOW_ICON_TOOLTIP):format(GetString(SI_PENTEST_OFF_BALANCE)),
			default = def.OffBalanceIcon,
			getFunc = function() return db.OffBalanceIcon end,
			setFunc = function(value)
						db.OffBalanceIcon = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_SHOW_ICON):format(GetString(SI_PENTEST_MINOR_VULNERABILITY)),
			tooltip = GetString(SI_PENTEST_MENU_SHOW_ICON_TOOLTIP):format(GetString(SI_PENTEST_MINOR_VULNERABILITY)),
			default = def.MinorVulnerabilityIcon,
			getFunc = function() return db.MinorVulnerabilityIcon end,
			setFunc = function(value)
						db.MinorVulnerabilityIcon = value
						UpdateLayout()
					  end,
		},
		{
			type = "checkbox",
			name = GetString(SI_PENTEST_MENU_INCOMBAT),
			tooltip = GetString(SI_PENTEST_MENU_INCOMBAT_TOOLTIP),
			default = def.onlyCombat,
			getFunc = function() return db.onlyCombat end,
			setFunc = function(value)
						db.onlyCombat = value
					  end,
		},
		{
			type = "slider",
			name = GetString(SI_PENTEST_MENU_MAXPEN),
			tooltip = GetString(SI_PENTEST_MENU_MAXPEN_TOOLTIP),
			min = 1,
			max = 18200,
			step = 1,
			default = def.maxpen,
			getFunc = function() return db.maxpen end,
			setFunc = function(value)
						db.maxpen = math.max(value, 1)
					  end,
		},
		{
			type = "editbox",
			name = GetString(SI_PENTEST_MENU_CRUSHER),
			tooltip = GetString(SI_PENTEST_MENU_CRUSHER_TOOLTIP),
			default = def.crusherValue,
			getFunc = function() return db.crusherValue end,
			setFunc = function(value)
					if value then

						local number = zo_round(tonumber(value) or def.crusherValue)
						db.crusherValue = number

					end
				end
		},
		{
			type = "editbox",
			name = GetString(SI_PENTEST_MENU_FORCEOFNATURE_CP),
			tooltip = GetString(SI_PENTEST_MENU_FORCEOFNATURE_CP_TOOLTIP),
			default = def.forceOfNatureValue,
			getFunc = function() return db.forceOfNatureValue end,
			setFunc = function(value)
					if value then

						local number = zo_round(tonumber(value) or def.forceOfNatureValue)
						db.forceOfNatureValue = number

					end
				end
		},
	}

	menu:RegisterOptionControls("PenTest_Options", options)

	CALLBACK_MANAGER:RegisterCallback("LAM-PanelOpened", ShowItems )
	CALLBACK_MANAGER:RegisterCallback("LAM-PanelClosed", ClearItems )

	return menu
end

-- default values (see http://wiki.esoui.com/AddOn_Quick_Questions#How_do_I_save_settings_on_the_local_machine.3F)
PenTest.defaults = {
	["tlw"] = {x = 150*dx, y = 150*dx, height = zo_round(25/dx)*dx, width = zo_round(150/dx)*dx},
	["iconsize"] = 25,
	["accountwide"] = true,
	["locked"] = false,
	["maxpen"] = 18200,
	["onlyCombat"] = false,
	["opacity"] = 100,
	["spellpen"] = true,
	["weaponpen"] = true,
	["CritDamageIcon"] = true,
	["OffBalanceIcon"] = true,
	["MinorVulnerabilityIcon"] = true,
	["crusherValue"] = 2108,
	["forceOfNatureValue"]= 3960,

}

-- Initialization
function PenTest:Initialize(event, addon)

	local name = self.name
	if addon ~= name then return end --Only run if this addon has been loaded

	-- load saved variables

	db = ZO_SavedVars:NewAccountWide(name.."_Save", 1, nil, self.defaults) -- taken from Aynatirs guide at http://www.esoui.com/forums/showthread.php?t=6442

	if db.accountwide == false then

		db = ZO_SavedVars:NewCharacterIdSettings(name.."_Save", 1, nil, self.defaults)
		db.accountwide = false

	end

	PenTest.db = db

	--register Events
	em:UnregisterForEvent(name.."load", EVENT_ADD_ON_LOADED)

	-- em:RegisterForEvent(name.."unit", EVENT_COMBAT_EVENT, onUnitDeath)
	-- em:AddFilterForEvent(name.."unit", EVENT_COMBAT_EVENT, REGISTER_FILTER_COMBAT_RESULT, 2260, REGISTER_FILTER_IS_ERROR, false) -- not needed?

	-- em:RegisterForEvent(name.."unit2", EVENT_COMBAT_EVENT, onUnitDeath)
	-- em:AddFilterForEvent(name.."unit2", EVENT_COMBAT_EVENT, REGISTER_FILTER_COMBAT_RESULT, 2262, REGISTER_FILTER_IS_ERROR, false)

	em:RegisterForEvent(name.."active", EVENT_PLAYER_ACTIVATED, GetAbilityData)
	em:RegisterForEvent(name.."combat", EVENT_PLAYER_COMBAT_STATE, onCombatState)
	em:RegisterForEvent(name.."target", EVENT_RETICLE_TARGET_CHANGED , onTargetChange)

	self.playername = zo_strformat("<<!aC:1>>",GetUnitName("player"))
	self.inCombat = IsUnitInCombat("player")

	self.MakeMenu()

	-- setup display frame

	tlw = PENTEST_TLW
	local dbtlw = db.tlw

	local isTLWShown = false
	local fragment = ZO_SimpleSceneFragment:New(tlw)

	function tlw.Toggle(self, show)

		show = show or not db.locked

		if show == isTLWShown then return end

		if show == true then

			HUD_SCENE:AddFragment(fragment)
			HUD_UI_SCENE:AddFragment(fragment)

		else

			HUD_SCENE:RemoveFragment(fragment)
			HUD_UI_SCENE:RemoveFragment(fragment)

		end

		isTLWShown = show
	end

	if (dbtlw) then

		tlw:ClearAnchors()
		tlw:SetAnchor(TOPLEFT, GuiRoot, TOPLEFT, dbtlw.x*dx, dbtlw.y*dx)

	end

	tlw:SetMovable(not db.locked)
	tlw:SetMouseEnabled(not db.locked)

	tlw:SetHandler("OnMoveStop", function(control)

		local x, y = control:GetScreenRect()

		dbtlw.x = zo_round(x/dx)
		dbtlw.y = zo_round(y/dx)

		tlw:ClearAnchors()
		tlw:SetAnchor(TOPLEFT, GuiRoot, TOPLEFT, dbtlw.x*dx, dbtlw.y*dx)
	end)

	-- Assign Textures:

	local tex1 = "esoui/art/icons/ability_buff_major_force.dds" -- Off Balance
	local tex2 = GetAbilityIcon(39077) -- Off Balance
	local tex3 = GetAbilityIcon(68359) -- Minor Vulnerabiltiy

	local iconControl = tlw:GetNamedChild("IconsControl")

	local icon1 = iconControl:GetNamedChild("CritDamageIcon")
	local icon1tex = icon1:GetNamedChild("Texture")
	icon1tex:SetTexture(tex1)
	icon1tex:SetAlpha(0.6)

	local icon2 = iconControl:GetNamedChild("OffBalanceIcon")
	icon2:GetNamedChild("Texture"):SetTexture(tex2)

	local icon3 = iconControl:GetNamedChild("VulnerabilityIcon")
	icon3:GetNamedChild("Texture"):SetTexture(tex3)

	UpdateLayout()

	-- setup display scenes

	GetAbilityData()
	onTargetChange()
	GetStaticValues()

end

-- Finally, we'll register our event handler function to be called when the proper event occurs.
em:RegisterForEvent(PenTest.name.."load", EVENT_ADD_ON_LOADED, function(...) PenTest:Initialize(...) end)